from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Label


class Arrows(Container):
    def compose(self) -> ComposeResult:
        stem_length = int(self.styles.width.value) - 1
        yield Label(stem_length * "─" + "▶", id="right")
        yield Label("◀" + stem_length * "─", id="left")



from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.events import DescendantBlur, Event, Mount
from textual.reactive import reactive
from textual.timer import Timer
from textual.validation import Integer
from textual.widgets import Button, Input, RichLog

from termchar_demo.buffer import Buffer


class Device(Container):
    class DataOut(Event):
        def __init__(self, data: str, sender: "Device") -> None:
            super().__init__()
            self.data = data
            self.sender = sender

    class DataIn(Event):
        def __init__(self, data: str) -> None:
            super().__init__()
            self.data = data

    class MessageRead(Event):
        def __init__(self, msg: str) -> None:
            super().__init__()
            self.msg = msg

    BINDINGS = [("ctrl+o", "focus_message", "Focus message")]

    BORDER_TITLE: str = "Device"
    TIMEOUT = 2

    is_busy_reading = reactive(False)
    timeout: int
    timeout_timer: Timer | None = None

    def output_widgets(self) -> ComposeResult:
        yield Input(id="output")
        with Horizontal():
            yield Input(id="write-termchars")
            yield Button("Write", id="write-button", variant="primary")

    def input_widgets(self) -> ComposeResult:
        with Horizontal():
            yield Buffer(id="input-buffer", length=30)
            yield Button("Clear", id="clear-button", variant="primary")
        with Horizontal():
            yield Input(
                placeholder="Type termination characters here...",
                id="read-termchars",
            )
            yield Input(
                value=str(self.TIMEOUT),
                id="timeout",
                restrict="[0-9]*",
                validators=Integer(minimum=0),
            )
            yield Button("Read", id="read-button", variant="primary")

    def log_widget(self) -> ComposeResult:
        yield RichLog(id="log", markup=True)

    def compose(self) -> ComposeResult:
        with Vertical(id="container"):
            yield from self.output_widgets()
            yield from self.log_widget()
            yield from self.input_widgets()

    def on_mount(self) -> None:
        self.query_one("#write-termchars").border_title = "Write Termination Characters"
        self.query_one("#log").border_title = "Application Log"
        self.query_one("#input-buffer").border_title = "Input Buffer"
        self.query_one("#read-termchars").border_title = "Read Termination Characters"
        self.query_one("#timeout").border_title = "Timeout"
        self.timeout = self.TIMEOUT

    def watch_is_busy_reading(self, is_busy_read: bool) -> None:
        def unlock_read():
            self.is_busy_reading = False

        if is_busy_read:
            output: Input = self.query_one("#output")
            output.disabled = True
            output.placeholder = "Waiting for message..."
            output.add_class("busy")
            self.query_one("#read-button").disabled = True
            self.query_one("#read-termchars").disabled = True
            self.query_one("#timeout").disabled = True
            self.timeout_timer = self.set_timer(
                delay=self.timeout, callback=unlock_read
            )
        else:
            output: Input = self.query_one("#output")
            output.disabled = False
            output.placeholder = "Type here to send data"
            output.remove_class("busy")
            self.query_one("#read-button").disabled = False
            self.query_one("#read-termchars").disabled = False
            self.query_one("#timeout").disabled = False
            if self.timeout_timer:
                self.timeout_timer.stop()

    @on(Input.Submitted, "#timeout")
    @on(DescendantBlur, "#timeout")
    def set_timeout(self) -> None:
        widget: Input
        if not (widget := self.query_one("#timeout")).is_valid:
            widget.value = "0"
        self.timeout = int(widget.value)

    @on(Input.Submitted, "#output")
    @on(Input.Submitted, "#write-termchars")
    @on(Button.Pressed, "#write-button")
    def write(self) -> None:
        termchars: Input = self.query_one("#write-termchars").value
        output: Input = self.query_one("#output")
        self.post_message(self.DataOut(output.value + termchars, sender=self))
        output.clear()

    @on(Input.Changed, "#read-termchars")
    def update_input_buffer(self, event: Input.Changed) -> None:
        self.query_one("#input-buffer").termchars = event.input.value

    @on(Input.Submitted, "#read-termchars")
    @on(Button.Pressed, "#read-button")
    def perform_read(self) -> None:
        self.read()

    def read(self) -> None:
        termchars: Input = self.query_one("#read-termchars").value
        msg = self.query_one("#input-buffer").read(termchars=termchars)
        if msg is not None:
            self.post_message(self.MessageRead(msg))
            self.is_busy_reading = False
        else:
            self.is_busy_reading = True

    @on(Button.Pressed, "#clear-button")
    def clear_buffer(self) -> None:
        self.query_one("#input-buffer").clear()

    @on(MessageRead)
    def log_read(self, event: MessageRead) -> None:
        self.query_one("#log").write(f'[dark_olive_green1]> Read ← "{event.msg}"')

    @on(DataOut)
    def log_write(self, event: DataOut) -> None:
        msg = event.data.removesuffix(self.query_one("#write-termchars").value)
        self.query_one("#log").write(f'[light_steel_blue1]> Write → "{msg}"')

    @on(DataIn)
    def fill_buffer(self, event: DataIn) -> None:
        buffer: Buffer = self.query_one("#input-buffer")
        buffer.append(event.data)
        if self.is_busy_reading:
            self.read()

    def action_focus_message(self) -> None:
        self.query_one("#output").focus()


class SimpleDevice(Device):
    TIMEOUT = 0

    def input_widgets(self) -> ComposeResult:
        with Horizontal():
            yield Buffer(id="input-buffer")
            yield Button("Clear", id="clear-button", variant="primary")
        with Horizontal():
            yield Input(
                placeholder="Type termination characters here...",
                id="read-termchars",
            )
            yield Button("Read", id="read-button", variant="primary")

    def compose(self) -> ComposeResult:
        with Vertical(id="container"):
            yield from self.output_widgets()
            yield from self.log_widget()
            yield from self.input_widgets()

    @on(Mount)
    def on_mount(self, event: Mount) -> None:
        event.prevent_default()
        self.query_one("#write-termchars").border_title = "Write Termination Characters"
        self.query_one("#log").border_title = "Application Log"
        self.query_one("#input-buffer").border_title = "Input"
        self.query_one("#read-termchars").border_title = "Read Termination Characters"

    def watch_is_busy_reading(self, is_busy_read: bool) -> None:
        if is_busy_read:
            output: Input = self.query_one("#output")
            output.disabled = True
            output.placeholder = "Sending auto-replies..."
            output.add_class("busy")
            self.query_one("#read-button").disabled = True
            self.query_one("#read-termchars").disabled = True
        else:
            output: Input = self.query_one("#output")
            output.disabled = False
            output.placeholder = "Type here to send data"
            output.remove_class("busy")
            self.query_one("#read-button").disabled = False
            self.query_one("#read-termchars").disabled = False



import asyncio
from random import choice

from textual import on, work
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Input

from termchar_demo.advanced_demo import AdvancedDemo, Client, Server
from termchar_demo.arrows import Arrows
from termchar_demo.devices import SimpleDevice

REPLIES = [
    "Got it, thanks!",
    "Received, thanks!",
    "Thanks, noted.",
    "Understood, thanks!",
    "Acknowledged.",
    "Thanks, will do.",
    "Got it.",
    "Noted.",
    "Copy that.",
    "Received.",
]

DELAY = 1.0


class BasicClient(Client, SimpleDevice):
    @on(Client.DataOut)
    def focus_write_button(self) -> None:
        self.query_one("#write-button").focus()

    def read(self) -> None:
        super().read()
        self.is_busy_reading = False


class BasicServer(Server, SimpleDevice):
    BORDER_TITLE: str = "Server"
    TIMEOUT = 0

    def on_mount(self) -> None:
        (widget := self.query_one("#read-termchars")).value = r"\n"
        widget.disabled = True
        (widget := self.query_one("#write-termchars")).value = r"\r\n"
        widget.disabled = True
        self.query_one("#read-button").disabled = True
        self.query_one("#write-button").disabled = True
        self.is_busy_reading = True

    def read(self) -> None:
        self.myworker()

    @work
    async def myworker(self):
        await asyncio.sleep(DELAY)
        termchars: Input = self.query_one("#read-termchars").value
        msg = self.query_one("#input-buffer").read(termchars=termchars)
        if msg is not None:
            self.post_message(self.MessageRead(msg))
            await asyncio.sleep(DELAY)
            termchars: Input = self.query_one("#write-termchars").value
            self.post_message(
                self.DataOut(
                    choice(REPLIES) + termchars,
                    sender=self,
                )
            )
            await asyncio.sleep(DELAY)
            self.query_one("#output").add_class("active")
            await asyncio.sleep(DELAY)
            self.query_one("#output").remove_class("active")


class BasicDemo(AdvancedDemo):
    def compose(self) -> ComposeResult:
        with Horizontal():
            yield BasicClient()
            yield Arrows()
            yield BasicServer()

    @on(SimpleDevice.DataOut)
    def send_data(self, event: SimpleDevice.DataOut) -> None:
        event.prevent_default()
        if event.sender == self.query_one(Client):
            target = Server
            dataflow_target = "#right"
        else:
            target = Client
            dataflow_target = "#left"
        self.animate_send_data(event.data, target, dataflow_target)

    @work
    async def animate_send_data(
        self, data: str, data_target: SimpleDevice, dataflow_target: str
    ) -> None:
        await asyncio.sleep(DELAY)
        arrow = self.query_one(dataflow_target).add_class("active")
        await asyncio.sleep(DELAY)
        arrow.remove_class("active")
        self.query_one(data_target).post_message(SimpleDevice.DataIn(data))



import importlib.resources

from textual import on
from textual.app import App, ComposeResult
from textual.widgets import Footer, Header, Markdown, TabbedContent, TabPane

from termchar_demo.advanced_demo import AdvancedDemo
from termchar_demo.basic_demo import BasicDemo


class TermCharDemo(App[None]):
    CSS_PATH = "app.tcss"

    BINDINGS = [
        ("ctrl+c", "quit", "Quit"),
    ]

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()
        with TabbedContent():
            with TabPane("Introduction", id="intro"):
                yield Markdown(
                    (
                        importlib.resources.files("termchar_demo") / "intro.md"
                    ).read_text()
                )
            with TabPane("Basic", id="basic"):
                yield BasicDemo()
            with TabPane("Advanced", id="advanced"):
                yield AdvancedDemo()
            with TabPane("Background", id="background"):
                yield Markdown(
                    (
                        importlib.resources.files("termchar_demo") / "background.md"
                    ).read_text()
                )

    @on(Markdown.LinkClicked)
    def open_web_link(self, event: Markdown.LinkClicked) -> None:
        self.open_url(event.href)

    def action_intro(self) -> None:
        self.query_one(TabbedContent).active = "intro"

    def action_basic(self) -> None:
        self.query_one(TabbedContent).active = "basic"

    def action_advanced(self) -> None:
        self.query_one(TabbedContent).active = "advanced"


def main():
    TermCharDemo().run()


app = TermCharDemo
if __name__ == "__main__":
    main()



from rich.text import Text
from textual.reactive import reactive
from textual.widgets import Label


class Buffer(Label):
    data = reactive("")
    termchars = reactive("")

    def __init__(
        self,
        data: str = "",
        length: int | None = None,
        termchars: str = "",
        *args,
        **kwargs,
    ) -> None:
        super().__init__(data, *args, **kwargs)
        self.length = length
        if length is not None:
            self.styles.max_width = self.length + 2
        self.data = data
        self.termchars = termchars

    def render(self) -> str:
        if not self.termchars:
            data = f"[red]{self.data}[/]"
        elif self.termchars in self.data:
            messages = self.data.split(self.termchars)
            if (termchars := self.termchars).endswith("\\"):
                # a backslash just before a markup tag cancels that tag
                # escape the backslash
                termchars += "\\"
            data = (
                "[green]"
                + f"[/][blue]{termchars}[/][green]".join(messages[:-1])
                + f"[/][blue]{termchars}[/][orange1]{messages[-1]}[/]"
            )
        else:
            data = f"[orange1]{self.data}[/]"
        # use non-breaking space to hide background
        data = data.replace(" ", " ")
        # truncate label with ellipsis if necessary
        text = Text.from_markup(data)
        text.truncate(self.size.width, overflow="ellipsis")
        return text

    def append(self, new_data: str) -> None:
        self.data = (self.data + new_data)[: self.length]

    def clear(self) -> None:
        self.data = ""

    def watch_data(self):
        if self.length is not None:
            fill_factor = len(self.data) / self.length
            if fill_factor >= 0.8:
                self.set_classes("full")
            elif fill_factor >= 0.6:
                self.set_classes("almost-full")
            else:
                self.set_classes("")

    def read(self, termchars: str) -> str | None:
        if not termchars:
            if self.data:
                data = self.data
                self.data = ""
                return data
            else:
                return None

        if termchars in self.data:
            msg, self.data = self.data.split(termchars, maxsplit=1)
            return msg
        else:
            return None



from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical

from termchar_demo.arrows import Arrows
from termchar_demo.devices import Device


class Client(Device):
    BORDER_TITLE: str = "Client"


class Server(Device):
    BORDER_TITLE: str = "Server"
    TIMEOUT = 0

    def compose(self) -> ComposeResult:
        with Vertical(id="container"):
            yield from self.input_widgets()
            yield from self.log_widget()
            yield from self.output_widgets()

    def on_mount(self) -> None:
        self.query_one("#read-termchars").value = r"\n"
        self.query_one("#write-termchars").value = r"\r\n"


class AdvancedDemo(Container):
    def compose(self) -> ComposeResult:
        with Horizontal():
            yield Client()
            yield Arrows()
            yield Server()

    @on(Device.DataOut)
    def send_data(self, event: Device.DataOut) -> None:
        if event.sender == self.query_one(Client):
            target = Server
        else:
            target = Client
        self.query_one(target).post_message(Device.DataIn(event.data))
